
/** Send and receive UDP packets to/from server.
 * nonblocking is set (and optionally udpbindport()) before first recvfrom()
 * gcc -Wall -o udpclient udpclient.c
 * ./udpclient serverip serverport 'get udp:4444'
 * Originally on linux.
 * Compiles,links,runs on windows (with cygwin).
 * Without cygwin ? */

#include <unistd.h>
#include <fcntl.h>
#include <error.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <arpa/inet.h>

#define BUFLEN 512
#define RCV_BUFLEN 1024
#define NPACK 10

/* name lookup not implemented, IP only: */
#define SRV_IP "10.8.21.252"
#define SRV_PORT 8078
/* tcpdump -n -v -i wlan1 udp port 8078 */

char *srv_ip;
int srv_port, message_len;
char buf[BUFLEN];
char rcv_buf[RCV_BUFLEN];

void diep(char *s)
{
   perror(s);
   exit(1);
}

/** bind to udpport
* on 0.0.0.0, so not to certain interface;
* Works well with server.
* But on client socket, only first call takes effect.
* Also, only effective before first recvfrom() ... after that the automatically assigned source port will stick */
int udpbindport(int sockfd, int udpport)
{
    struct sockaddr_in servaddr;
    bzero(&servaddr, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr=htonl(INADDR_ANY);
    servaddr.sin_port=htons( udpport );
    return bind(sockfd,(struct sockaddr *)&servaddr, sizeof(servaddr));
}

/** ./udpclient serverip serverport 'get udp:4444'
*/
int main(int argc, char**argv)
{
  struct sockaddr_in si_other;
  int s, i, n, rcvd_cnt;
  socklen_t slen=sizeof(si_other);

  srv_ip = argc > 1 ? argv[1] : SRV_IP; /* process commandline arguments */
  if( argc <=2 || ((srv_port = atoi(argv[2])) == 0 )){ /* no port argument, or bogus */
      srv_port = SRV_PORT;
  }
  if( argc > 3 ){ /* message specified on commandline, arg 3 */
      char *message = argv[3];
      message_len = strlen( message );
      if( message_len > BUFLEN ) message_len = BUFLEN;
      memcpy( buf, message, message_len); /* strncpy would look cleaner, but costs more, and we don't need padding */
  } else { /* message not specified: example content */
      sprintf(buf, "This is an UDP packet \n");
      message_len = strlen(buf); // or BUFLEN to send longer packets;
  }

  if ((s=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP))==-1)
    diep("socket");

  memset((char *) &si_other, 0, sizeof(si_other));
  si_other.sin_family = AF_INET;
  si_other.sin_port = htons(srv_port);
  if (inet_aton(srv_ip, &si_other.sin_addr)==0) {
    fprintf(stderr, "inet_aton() failed\n");
    exit(1);
  }

  rcvd_cnt = 0;
  for (i=0; i<NPACK; i++) {	/* sends very fast, no delay... tcpdump shows 7..10 usec between packets */
    if( i == NPACK-2 ){ /* emulate packet drop: after some good packets, intentionally SCREW up the request so we do NOT get response from server... */
      buf[0] = 'Z';
      printf("Sending SCREWED packet %d\n", i);
    } else { /* normally */
      printf("Sending packet %d\n", i);
    }
    if( i == 3) udpbindport(s, 2000+i); /* normally NOT necessary for a client, but we experiment. Only effective with i == 0, before first recvfrom() */
    if (sendto(s, buf, message_len, 0, (struct sockaddr *)&si_other, slen)==-1)
      diep("sendto()");

    int flags = fcntl(s, F_GETFL, 0);
    if ( i == 0 && (n=fcntl(s, F_SETFL, flags | O_NONBLOCK)) ){ /* after some blocking ops, set to nonblocking; otherwise recvfrom() blocking when response is lost */
      printf( "ERROR: fcntl(s, F_SETFL, flags | O_NONBLOCK) returned %d\n", n);
    }
    /* first packet arrives from server after 1-23 msec (23 msec is only after inactivity; possibly caused by wifi-state); than ~120 usec between packets */
    n = recvfrom(s, rcv_buf, RCV_BUFLEN, 0, (struct sockaddr *)&si_other, &slen);
    if( n != 0 ){ // EAGAIN=%d EWOULDBLOCK=%d errno=%d, ; EAGAIN, EACCES, errno,
      if( n > 0 ){
	rcvd_cnt ++;
      }
      printf( "Received packet (%d): %s\n", n, (n > 0) ? rcv_buf : "ERR");
    }
  }
  
  if( rcvd_cnt < NPACK ){ /* more receive attempts after some delay */
    int more_attempts = NPACK - rcvd_cnt;
    sleep(1);
  
    while(more_attempts--){
	n = recvfrom(s, rcv_buf, RCV_BUFLEN, 0, (struct sockaddr *)&si_other, &slen);
	if( n != 0 ){
	  if( n > 0 ){
	    rcvd_cnt ++;
	  }
	  printf( "RECEIVED packet (%d) (rcvd_cnt=%d): %s\n", n, rcvd_cnt, (n > 0) ? rcv_buf : "ERR");
	}
    }
    /* we've received all response packets; note: the SCREWED packet that didn't generate response.
     In real life, the 1 sec might NOT be enough in some situations => more attempts needed if no answer
     */
  }
  close(s);
  return 0;
}
