
/** vemsfriend.c 2014-12-21 (last modified: 2014-12-28 16:27)
* Simple and lightweight UDP rendezvous-server
* "Public Meeting point" when both endpoints are behind NAT (just at the start of communication, not relaying all data => best possible VOIP latency).
* Peers can find each others before UDP hole-punching.
*
* The problem of data-connection between peers both behind NAT is more significant every year, negatively effecting Tor, torrents, P2P and F2F networks.
* Our civilization lost at least 1300 years at least once
* (the creation method of money, and the lack of independently maintained knowledge-bases contributed to it:
* see The Money Masters http://www.themoneymasters.com/the-money-masters/ and  http://en.wikipedia.org/wiki/Library_of_Alexandria )
* Knowledge bases maintained and shared over F2F networks should be used extensively to maintain and develop knowledge under harsh conditions.
* Works on most type of NATs. But depends on UDP (TCP connection directly - without central relay - usually not possible when both endpoints are behind NAT;
* This method works almost everywhere, except some large corporate networks that don't support UDP : httptunnel hts might be useful there, outside the scope now).
*
* Works for an arbitrary number of node-pairs, not limited by dyndns accounts (or adversary attacks effecting dyndns service).
* Also works if UPNP is not supported on the NAT device of any endpoint.
* 
* written by: Marcell Gal
* When 2 peers, both with private IP address behind different NATs want to communicate directly
* (for VOIP, udptunnel ssh, or "VPN", or establish channel(s) for data-sharing RetroShare / torrent / ...):
* they can find out the partner's IP (and their own also) if they provide the same key.
*
* Traditionally this communication was achieved with dynamic DNS and UPNP.
* DNS is an unnecessary centralized technology involved in internet communications, a potential (and real !) subject of adversary attacks
* (People should remember the 2014-06-30 Microsoft attack on noip.org,
* that directly effected millions of DNS entries, and indirectly tens of millions of people for several days, first few days of July 2014,
* causing tens to hundred millions of dollars damage, 
* actually caused a few 100 EUR damage just to me, could have been thousands had I been unable to act quickly, eg. if on a vacation) :
* Causing damage to its users (and to those who don't use Microsoft at all) seems to be the rule for Microsoft, not the exception).
* https://www.eff.org/deeplinks/2014/07/microsoft-and-noip-what-were-they-thinking
* http://www.noip.com/blog/2014/07/10/microsoft-takedown-details-updates/
*
* In the long run, "advertisements" should be signed by the issuer: keyserver DHT network would be a reasonable alternative to the DNS which is easily abused through central-authority.
* In the short term, the existing keyserver network could be used http://en.wikipedia.org/wiki/Key_server_%28cryptographic%29 if it can (be modified to) scale sufficiently.
*
* Even when dyndns works (not actually brought down; terms of service not changed; enough dyndns accounts; bills paid and not forgotten),
* user is often not authorized to configure UPNP on the NAT box
* (eg. in an internet cafe, small office, public hotspot, mobile broadband, or the cable service provider's NAT box).
*
* This is just an initial prove-of-concept to allow peers both behind NAT to find each other, and build data-bridges.
* STUN server lacks the feature of looking up "peer by key": http://en.wikipedia.org/wiki/STUN
*
* Possible improvements later:
* IMPORTANT: larger keyspace (at least 20 or 32 bytes for sha1/sha256 hashes);
* anti-spoofing measures (server challenges client before updating timestamp / IP records).
* cryptographic authentication
*	(otherwise an agressive sniffing and source-spoofing adversary can make it much slower for the honest peers to find and ssl / ssh authenticate each other;
*	although if an adversary can sniff the packets of all N vemsfriend servers used by peers,
*	it might be close to one endpoint and possibly able to alter packets and prevent comm anyway - eg. cable provider)
* hashcash (helps prevent DOS attacks: client proof of work: possibly some pooled LTC mining)
* statistics
* usage examples (eg. patched chownat.pl)
* distributed through naive client (attempt to contact N servers until success; success if any of them is alive)
* distributed through data propagation (DHT service)
* Persistancy: tokyocabinet
* Payment processing (guaranteed service: long term data storage for those who pay; including graph and "wallet" date, similar to stellar.org).
* Relaying UDP data (to allow mapping of network connectivity)
* Tor hidden service (much slower, but has advantages; the original UDP method is lightning fast)
* Tor UDP alternative would be awesome (the problem of connection between peers both behind NAT is growing every year).
* Simple hash or critbit-tree implementation over TCP.
* Note: socat is an awesome program, and might be a better bed for the last few (socat already relays TCP and SSL data). 
* ----
* For now just find each other before establishing communication (typically over UDP like udptunnel or chownat.pl).
* Written in the hope that it might be useful for civilization, eg. to help people learn useful ways for the creation of money, which would be an achievement comparable (in usefulness) to the internet.
* Might be very useful to integrate in RetroShare as an alternative way of "bootstrapping"
* (if configured so, or when UPNP fails: UPNP depends on a certain NAT device often not administrable by the user, while this might work if any of N services is accessible;
* => should be possible to maintain enough servers to reach 99.999% successrate).
* Originally used with http://samy.pl/chownat/
* With chownat.pl the server MUST also know the client IP address (and UDP port).
* Note: not needed with pwnat, but pwnat usually does not work in real life (some firewalls don't like the otherwise smart and harmless ICMP TTL_EXCEEDED "impersonation").
* TODO: implement hash so peers can use the same (ASCII) key and find each other. Either tokyocabinet, or a simpler in-memory hash.
*/

#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <arpa/inet.h>

#define UDB_BUFLEN 1000
/* at least +15 bytes at the end for the IP */
char mesg[UDB_BUFLEN+24];
const char *srv_public_ip;
time_t now;	/* actual time(); make sure recently updated before use */

#define UDP_PORT_DEFAULT 32000

#define PAIR_ENTRIES 70000

/* change this to some preshared-key (sent over the wire in plaintext) */
#define REQ_STR "get udp:"
/* length of this string: */
#define REQ_SLEN 8
/* get udp:1 .. get_udp:99999  ... there is normally a \n (or \r\n) at the end, so +2 .. +7 instead of +1 .. +5 */
#define REQ_MINLEN REQ_SLEN+2
#define REQ_MAXLEN REQ_SLEN+7

#define IP_LEN sizeof(struct in_addr)

struct udp_peer_t
{
    struct in_addr ipaddr; /* IP_addr */
    time_t timestamp;	/* last seen */
};

struct pair_t
{
  struct udp_peer_t ab[2];
};

struct pair_t pairs[PAIR_ENTRIES];

/* negative means error */
int check_request(char *buf, int len)
{
  int ret;
  if(len < REQ_MINLEN) return -1;
  if(len > REQ_MAXLEN) return -2;
  // compare REQ_SLEN characters, and see if buf starts with REQ_STR...
  if( strncmp(REQ_STR, buf, REQ_SLEN) != 0) return -3; /* BAD greeting */
    
  ret = atoi(buf + REQ_SLEN); /* note: also recognizes negative numbers */
  return ret;
}

/** return 0 if same IP */
int store_ip()
{
  return 0;
}

/** Update timestamp, or store ip (replacing oldest "Least Recently Seen" entry)
 * return the pair (possibly 0.0.0.0 if no pair yet), 
 * idx must be checked by caller 0 .. PAIR_ENTRIES-1 is acceptable
 * now is updated with actual time() */
struct udp_peer_t* find_pair(struct in_addr *ip, int idx)
{
  int sti = 0; /* store in [1] if not  */
  time_t t0, t1;
  int diff;
  struct pair_t *p = &pairs[idx];
  
  now = time(NULL);
  if( memcmp( ip, &p->ab[0].ipaddr, IP_LEN ) == 0){ /* found */
      p->ab[ 0 ].timestamp = now; /* update timestamp */
      return &p->ab[1]; /* return the pair: OTHER idx */
  }

  /* ip not in [0], Maybe in [1] ? Let's see: */
  if( memcmp( ip, &p->ab[1].ipaddr, IP_LEN ) == 0){ /* found */
      p->ab[ 1 ].timestamp = now; /* update timestamp */
      return &p->ab[0]; /* return the pair: OTHER idx */
  }

  /* not found in [0] nor [1]. Find the older timestamp and overwrite that (do NOT check against 0.0.0.0 because that's only true initially !) */
  t1 = p->ab[1].timestamp; /* if either t0 or t1 is uninitialized 0, store there */
  t0 = p->ab[0].timestamp;
  if( t0 == 0 ){ /* could be written more efficiently, but easier to understand this way */
      sti = 0;
  } else if( t1 == 0 || ((diff = t1 - t0) < 0) ){ /* t1 uninitialized or t1 was older: perhaps it was 0, or some alien or old ip: will be overwritten */
      sti = 1;
  }
#ifdef DEBUG
  char *str;
  str = inet_ntoa(p->ab[0].ipaddr);
  printf("WILL OVERWRITE (%d)(%s, %ld ", sti, str, p->ab[0].timestamp );
  str = inet_ntoa(p->ab[1].ipaddr); /* insiduous: never inet_ntoa() before printing or copying the result of prev call */
  printf("; %s, %ld) LEN=%d\n", str, p->ab[1].timestamp, IP_LEN);
#endif

  memcpy(&p->ab[ sti ].ipaddr, ip, IP_LEN );
  p->ab[ sti ].timestamp = now; /* store timestamp also */

#ifdef DEBUG
  str = inet_ntoa(p->ab[0].ipaddr);
  printf("AFTER OVERWRITE (%s, %ld", str, p->ab[0].timestamp );
  str = inet_ntoa(p->ab[1].ipaddr);
  printf("; %s, %ld)\n", str, p->ab[1].timestamp );
#endif

  return &p->ab[sti ^ 1]; /* return the pair: OTHER idx */
}

/** append str to buf; Return the number of characters appended.
* Caller must make sure result will fit, eg. str=inet_ntoa() length is limited by nature */
int str_append(char *buf, const char *str)
{
      int slen;
      slen = strlen(str);
      strcpy(buf, str); /* appended */
      return slen;
}

/** append ip formatted as x.x.x.x to buf. Return the number of characters appended.
* special case: if the vemsfriend server and one endpoint are BOTH behind same NAT,
* the NAT box' internal 10.x.x.x IP address (should be configurable also) is seen - it's nonsense to advertise that */
int ip_append(struct in_addr *ip, char * buf )
{
  const char *str;
  int slen;
  str = inet_ntoa(*ip);  /* inet_ntop(AF_INET, &(((struct sockaddr_in *)sa)->sin_addr), s, maxlen); deprecated but simpler: */
  if( strncmp(str, "10.8.21.251", 11) == 0) {	/* special case: the request came from the same private subnet as this server => apply hack: */
      str = srv_public_ip;
  }
  slen = str_append( buf, str );
  return slen; 
}

/** append a space, and a decimal number "port" (which can be a time value) to buf.
 * Return the number of characters appended. */
int port_append(char * buf, time_t port )
{
  int slen;
  sprintf(buf, " %ld", port); /* more efficient method probably exists */
  slen = strlen( buf );
  return slen; 
}

/** bind to udpport
* on 0.0.0.0, so not to certain interface */
int udpbindport(int sockfd, int udpport)
{
    struct sockaddr_in servaddr;
    bzero(&servaddr, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr=htonl(INADDR_ANY);
    servaddr.sin_port=htons( udpport );
    return bind(sockfd,(struct sockaddr *)&servaddr, sizeof(servaddr));
}

/** gcc -Wall -o vemsfriend vemsfriend.c
 * ./vemsfriend 32001
 * echo 'get udp:65432' | nc -u <vemsfriendserverip> 32001
 * or rather
 * echo 'get udp:65432' | socat  - udp:<vemsfriendserverip>:32001
 * In the response (TODO: cleanup when response syntax stabilizes), first IP is ourselves; 2nd IP is our partner:
 * | awk -F '.' 'NF==4 {if(++CNT == 2){ print ; exit } }'  looks like the 2nd IP address
 * | gawk 'NF {if(++CNT==3){print $1; exit 0;}}'   prints 3d line 1st field
 * | tail -1   also works, because currently last line: but this might not be future-compatible
 * | awk 'END{print $1}' only prints first field (eg. not timestamp; which could be useful after the IP)
 */
int main(int argc, char**argv)
{
   int sockfd, n, udpport;
   struct sockaddr_in cliaddr;
   struct udp_peer_t *p;
   socklen_t len;

   /* our public IP can be provided as 2nd argument: */
   srv_public_ip = ( argc > 2) ? argv[2] : "80.98.222.153";	/* TODO: look up automatically ? From STUN ? Or via other server ? */
   sockfd=socket(AF_INET,SOCK_DGRAM,0);

   udpport = argc < 2 ? UDP_PORT_DEFAULT : atoi( argv[1] );
   udpbindport(sockfd, udpport);

   for (;;){
      struct in_addr *ip;
      len = sizeof(cliaddr);
      n = recvfrom(sockfd, mesg, UDB_BUFLEN, 0, (struct sockaddr *)&cliaddr, &len);
      int ret = check_request( mesg, n );
      if(( ret < 0 ) || ( ret >= PAIR_ENTRIES)) continue; /* if not correct: stop processing immediately don't send response, don't even LOG */ 
      
      /* would be nice to see which is the server... so server IP would NOT be purged from memory, and multiple clients could connect */
      ip = &cliaddr.sin_addr; /* remote IP */
      p = find_pair(ip, ret ); /* ret is the numeric "key" extracted from the message */
      mesg[n++] = '\n';
      n += ip_append( ip, mesg + n );  /* line2: <client IP> <current time> <client port>  ... also replaces iplocation service "STUN" */
      n += port_append( mesg + n, now ); /* current time updated in find_pair() */
      n += port_append( mesg + n, ntohs(cliaddr.sin_port) );

      mesg[n++] = '\n'; /* should we check if already contains newline ? */
      n += ip_append( &p->ipaddr, mesg + n ); /* line3: pair IP */
      n += port_append( mesg + n, p->timestamp );
      /* other partner's UDP port not sent: no usefulness */

      mesg[n++] = '\n';
      mesg[n] = 0; /* terminate string; n NOT incremented: terminating 0 not sent to peer */    
      printf("\n------------------\nReceived data. (code=%d) %s", ret, mesg);
      sendto(sockfd, mesg, n, 0, (struct sockaddr *)&cliaddr, sizeof(cliaddr)); /* send response */
   }
}
