
/** Send and receive UDP packets to/from server.
 * nonblocking is set (and optionally udpbindport()) before first recvfrom()
 * Used from pwnat (cd pwnat-orig; patch -p1 < pwnat.diff)
 * 
 * Or as a standalone program:
 * gcc -Wall -DMAIN -o vfclient vfclient.c
 * ./vfclient serverip serverport 'get udp:4321'
 * similar effect as:
 * PH=`echo 'get udp:4321' | socat - udp:$vfserver:$vfport | gawk 'NF {if(++CNT==3){print $1; exit 0;}}' `
 * pwnat -c ... $PH ...
 *
 * Originally on linux.
 * Compiles,links,runs on windows (with cygwin).
 * Without cygwin ? */

#include <unistd.h>
#include <fcntl.h>
#include <error.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <arpa/inet.h>
#include <netdb.h> 

#include "vfclient.h"

#define BUFLEN 512
#define RCV_BUFLEN 1024
#define NPACK 10

/* name lookup not implemented, IP only: */
#define SRV_IP "10.8.21.252"
#define SRV_PORT 8078
/* tcpdump -n -v -i wlan1 udp port 8078 */
#ifndef PARTY_KEY
#define PARTY_KEY 4321
#endif

const char *server_hostname;
char *client_ip; /* partner IP */
int vfresponse_len;	/* length of response received from vemsfriend server */
int srv_port, message_len, vf_sock;
char buf[BUFLEN];
char rcv_buf[RCV_BUFLEN];
struct hostent *server;
struct sockaddr_in si_vfserver;
int vfserver_flags, rcvd_cnt;
socklen_t slen=sizeof(si_vfserver);

void diep(const char *s)
{
   perror(s);
   exit(1);
}

/** bind to udpport
* on 0.0.0.0, so not to certain interface;
* Works well with server.
* But on client socket, only first call takes effect.
* Also, only effective before first recvfrom() ... after that the automatically assigned source port will stick */
int udpbindport(int sockfd, int udpport)
{
    struct sockaddr_in servaddr;
    bzero(&servaddr, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr=htonl(INADDR_ANY);
    servaddr.sin_port=htons( udpport );
    return bind(sockfd,(struct sockaddr *)&servaddr, sizeof(servaddr));
}

/** creates UDP IPV4 socket
 * initializes buf*/
int create_vf_sock(void)
{
  if((vf_sock=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0){
    diep("socket");
  }
  if(buf[0]==0){ /* message not specified: example content */
      sprintf(buf, "get udp:%d", PARTY_KEY);
      message_len = strlen(buf); // or BUFLEN to send longer packets;
  }
  return vf_sock;
}

/** vfserver_hostname:
 * eg. vfserver.domain.com or vfserver.domain.com:5555 (port)
 * return NULL if no success */
struct sockaddr_in* vfserver_set(const char * vfserver_hostname )
{
  char *port_str;
  int port;

  port_str = strchr(vfserver_hostname, ':');
  if( port_str && strlen(port_str) > 1 && ((port = atoi(port_str+1)) != 0 )){ /* :port specified */
      srv_port = port;
      printf("vfserver_hostname=%s\n", vfserver_hostname);
      *port_str=0; /* terminate with 0 so vfserver_hostname is only servername now */
      printf("vfserver_hostname=%s\n", vfserver_hostname);
  }

  memset((char *) &si_vfserver, 0, sizeof(si_vfserver));
  si_vfserver.sin_family = AF_INET;
  si_vfserver.sin_port = htons(srv_port);

  server = gethostbyname(vfserver_hostname);	/* gethostbyname: get the server's DNS entry */
  if (server == NULL) {
      fprintf(stderr,"ERROR, no such host as %s\n", vfserver_hostname);
      return NULL;
  }
  memcpy( &si_vfserver.sin_addr.s_addr, server->h_addr, server->h_length );

#if 0
  if (inet_aton(srv_ip, &si_vfserver.sin_addr)==0) { /* inet_aton() works with IP only */
    fprintf(stderr, "inet_aton() failed\n");
    exit(1);
  }
#endif
  return &si_vfserver;
}

void vf_sock_nonblock(void)
{
    int n;
    if( (vfserver_flags & O_NONBLOCK) == 0 ){ /* only do it once, initially */
      vfserver_flags = fcntl(vf_sock, F_GETFL, 0);
      vfserver_flags |= O_NONBLOCK ;
      n=fcntl(vf_sock, F_SETFL, vfserver_flags );

      if( n ){ /* set to nonblocking; otherwise recvfrom() blocks forever when response is lost */
	printf( "ERROR: fcntl(s, F_SETFL, vfserver_flags | O_NONBLOCK) returned %d\n", n);
      }
    }
}

/** find the 3d line, first word (line2 is self IP).
* srv_msg is vfresponse_len bytes long
* For 0.0.0.0 returns NULL currently */
char *vf_extract_partnerip( char *srv_msg )
{
  int line2; /* index of line2, than line3 */
  int cnt0, cntd, idx = 0;
  while( idx < vfresponse_len ){
    if( srv_msg[idx] == '\n' ) break;
    idx++;
  }
  line2 = ++idx;
  while( idx < vfresponse_len ){
    if( srv_msg[idx] == '\n' ) break;
    idx++;
  }
  line2 = ++idx; /* now line3 */
  /* if(idx < vfresponse_len) printf("line3=%d\n", line2); // debug only; don't enable in production without extra checks, overindexing */ 
  cnt0 = cntd = 0;
  while( idx < vfresponse_len ){
    char c = srv_msg[idx];
    if( c >= '0' && c <= '9' ){
      cntd++; /* count digits */
      if( c == '0') cnt0++; /* count zeroes */
    } else if( c == '\t' || c == ' ' || c == '\n' || c == '\r' ){ /* end of first word */
      srv_msg[idx] = 0; /* terminate ... change message, sorry */
      break;
    }
    idx++;
  }
  /* return NULL if overindexing; or if "IP" contains less than 4 digits, or if all digits are 0 */
  return ( line2 >= vfresponse_len || idx > vfresponse_len) || cntd < 4 || (cnt0 == cntd) ? NULL : (srv_msg+line2);
}

char *receive_vfresponse(int ntimes)
{
    int n, rcv_attempts = ntimes;
    rcvd_cnt = 0;
    do{ /* attempt receiving ... a few times. usleep would be useful */
      /* first packet arrives from server after 1-23 msec (23 msec is only after inactivity; possibly caused by wifi-state); than ~120 usec between packets */
	n = recvfrom(vf_sock, rcv_buf, RCV_BUFLEN, 0, (struct sockaddr *)&si_vfserver, &slen);
	if( n != 0 ){ // EAGAIN=%d EWOULDBLOCK=%d errno=%d, ; EAGAIN, EACCES, errno,
	  if( n > 0 ){
	    vfresponse_len = n;
	    rcvd_cnt ++;
	  }
	  // if(n>0) fprintf(stderr, "Received packet (%d): %s\n", n, (n > 0) ? rcv_buf : "ERR");
	  break;
	}
      } while( rcvd_cnt == 0 && rcv_attempts-- > 0);
      return rcvd_cnt ? rcv_buf : NULL;
}

/** call create_vf_sock() and vfserver_set() before this */
void send_vfquery(void)
{
    if (sendto(vf_sock, buf, message_len, 0, (struct sockaddr *)&si_vfserver, slen)==-1)
      diep("sendto()");
}

#ifdef MAIN
/** ./vfclient serverip serverport 'get udp:4444'
*/
int main(int argc, char**argv)
{
  int i;
   /* argument processing */
  server_hostname = argc > 1 ? argv[1] : SRV_IP; /* process commandline arguments */
  if( argc <=2 || ((srv_port = atoi(argv[2])) == 0 )){ /* no port argument, or bogus */
      srv_port = SRV_PORT;
  }
  if( argc > 3 ){ /* message specified on commandline, arg 3 */
      char *message = argv[3];
      message_len = strlen( message );
      if( message_len > BUFLEN ) message_len = BUFLEN;
      memcpy( buf, message, message_len); /* strncpy would look cleaner, but costs more, and we don't need padding */
  }
  create_vf_sock();
  vfserver_set( server_hostname );
  vf_sock_nonblock();
  /* udpbindport(vf_sock, 2000+i); // normally NOT necessary for a client. Important: Only effective with i == 0, before first recvfrom() */
  rcvd_cnt = 0;
  for (i=0; i < NPACK && rcvd_cnt == 0; i++){	/* when sending without delay: tcpdump shows 7..10 usec between packets */
    fprintf(stderr, "Sending UDP packet to vfserver\n");
    send_vfquery();
    client_ip = receive_vfresponse(5); /* big message at this point, not just client_ip */
    if( client_ip ){
      char * parner_ip_extracted = vf_extract_partnerip( client_ip );
      if( parner_ip_extracted ){
	fprintf(stderr, "partner_ip_extracted=%s\n", parner_ip_extracted);
	printf("%s\n", parner_ip_extracted);
      }
    }
    if( rcvd_cnt == 0 ){ /* 300, 500 msec would be more appropriate for first 2 retries */
      sleep(1);
    }
  }
  close(vf_sock);
  return 0;
}
#endif /* MAIN */

/** returns 0 if not starts with autovf: */
char * detect_autovf(char * srvstr)
{
  char *srv_hn = NULL;
  printf("detect_autovf(): testing if %s startswith autovf:\n", srvstr );
  if( strncmp(srvstr, "autovf:", 7 ) == 0 ){ /* startswith("autovf:") */
      srv_hn = srvstr + 7; /* server hostname */
      printf( "autovf mode:%s\n", srv_hn);
      // vf_sock = ...;
  }
  
  return srv_hn;
}
